# TODO: if included in field_name as part and multiple values by 's', 'es' at the end
#   like {'field': 'languages', 'generator': '[choice(languages), choice(languages)]'}
from typing import Optional, List
from pydantic import BaseModel


class Rule(BaseModel):
    field: str
    generator: str
    len: Optional[int] = None


class RuleSet(BaseModel):
    name: str
    rules: List[Rule]

# todo: move rules to pydantic models & define sets


person_rules = [
    {'field': 'user_id', 'generator': 'choice(["stone", "crosher", "valensia", "sun56", "bigmommy"]).lower()'}]


person_rules = RuleSet(name='person', rules=[Rule(**rule) for rule in person_rules])

default_rules = {
                 'user_id': {'generator': 'text.word().lower() + str(randint(10, 99))'},
                 'id': {'generator': 'uuid4().hex', 'len': 12},
                 'family_name': {'generator': 'person.surname()'},
                 'balance': {'generator': 'randint(350, 5500)'},
                 'amount': {'generator': 'randint(5, 1200) * choice([-1, 1])'},
                 'address': {'generator': 'address.address()'},
                 'default': {'generator': 'uuid4().hex', 'len': 6},
                 'rate': {'generator': 'randint(1, 5)/10'},
                 'login': {'generator': 'text.word()'},
                 'text': {'generator': 'text.text()'},
                 'nickname': {'generator': 'text.word()'},
                 'phone': {'generator': 'person.telephone()'},
                 'price': {'generator': 'round(random(), 2)'},
                 'ts': {'generator': 'current_time'},
                 'flag': {'generator': 'choice([\'Y\', \'N\'])'},
                 'date': {'generator': 'date.date(2019, 2021)'},
                 'city': {'generator': 'address.city()'},
                 'country_code': {'generator': 'choice(countries.codes)'},
                 'country': {'generator': 'choice(countries.names)'},
                 'language': {'generator': 'choice(languages)'},
                 'title': {'generator': 'text.text()', 'len': 20},
                 'surname': {'generator': 'person.surname()'},
                 'name': {'generator': 'person.name().split()[0]'},
                 'description': {'generator': 'text.text()'},
                 'email': {'generator': 'person.email()'}
                 }



""" main module that contains RunGenerator class that used to init data generation """
import os
import inspect
from collections import defaultdict
from typing import List, Dict, Text, Union, Tuple, Optional
from fakeme.tables import MultiTableRunner
from fakeme.fields import FieldRulesExtractor, FieldRules
from fakeme.config import Config
from fakeme.utils import log
from fakeme import config
from datetime import datetime

created = []


class FakemeException(Exception):
    pass


class Fakeme:
    def __init__(self,
                 tables: Union[List, Tuple[str, Union[str, dict, list]]],
                 with_data: List = None,
                 settings: Dict = None,
                 dump_schema: bool = True,
                 rls: Dict = None,
                 appends: List = None,
                 cli_path: Text = None,
                 field_rules: Optional[List] = None):
        """
        Main Class to start data generation

        :param tables: list with tuples, what contain
            (dataset/database (optional), table_id/table_name, path_to_table_schema
            or schema in one of supported formats), example:

                [
                    ("yourdatabase", "table_name1", "schemas/table_name1.json"),
                    ("yourdatabase", "table_name2", schema_in_python_object),
                    ("table_name3", "schemas/your_schema_in_json.json") ...
                ]

                where
                        schema_in_python_object = [
                                {
                                  "type": "STRING",
                                  "name": "field1",
                                  "mode": "NULLABLE"
                                 },
                                 {
                                  "type": "STRING",
                                  "name": "field2"
                                 }]
        :type tables: list
        :param settings:
            Example:
                    settings = {
                        'table_id' : {
                            'field': {
                                'unique_values': count_of_unique_values_in_column,
                                'unique': True
                            }
                        }
                    }
        :type settings: dict
        :param dump_schema:
        :type dump_schema:
        :param rls:
        :type rls:
        """
        if not field_rules:
            field_rules = []
        FieldRules.user_rules = field_rules
        self.tables = self.normalise_tables(tables)
        if not settings:
            settings = {}
        config.cfg = Config(**settings)
        self.cfg = config.cfg
        self.cfg.path_prefix = cli_path or os.path.dirname(inspect.stack()[1][1])
        self.rls = rls if rls else None or {}
        self.dump_schema = dump_schema
        # todo: remove this things with paths
        if not cli_path and os.getcwd() not in self.cfg.path_prefix:
            self.cfg.path_prefix = os.path.join(os.getcwd(), self.cfg.path_prefix)
        self.appends = appends or []
        self.cli_path = cli_path
        self.with_data = self.validate_data_source(with_data)

    @staticmethod
    def normalise_tables(tables):
        if (isinstance(tables, Tuple) or isinstance(tables, List)) and isinstance(tables[0], str):
            # mean we have only one table definition
            tables = tuple(tables,)
            return [tables]
        elif isinstance(tables, List):
            for table in tables:
                if not isinstance(table[0], str):
                    raise FakemeException(
                        f"Tables definitions must be presented like a tuple ('table_name', schema). "
                        f"You provided: {table}")
            return tables

    def validate_data_source(self, paths_list):
        if not paths_list:
            paths_list = []
        validated_list = []
        for path in paths_list:
            if not os.path.isfile(path):
                if self.cfg.path_prefix not in path:
                    path = os.path.join(self.cfg.path_prefix, path)
                    if not os.path.isfile(path):
                        raise Exception(f'Could not find the path {path} from "with_data" parameter')
            validated_list.append(path)
        return set(validated_list)

    def run(self):
        """
            call method to run data generation
            # todo: rewrite priority description
            priority 0 - tables that target for relations (
            priority 1 - tables that target for relations and in relations
            priority 2 - tables what not in self.append
            priority 3 - tables in self.append

        """
        log.info("Fakeme starts at", datetime.now())
        # get all fields and schemas for tables
        self.schemas, fields = MultiTableRunner(
            self.tables, rls=self.rls).\
            get_fields_and_schemas(dump_schema=self.dump_schema)
        if self.cfg.auto_alias:
            self.resolve_auto_alises()
        priority_dict = self.create_tables_priority_graph()
        field_extractor = FieldRulesExtractor(fields)
        # generate "value_rules.json" with rules for fields
        field_extractor.generate_rules()

        self.linked_fields = field_extractor.get_chains(self.schemas, self.rls)

        for level in priority_dict:
            for table in priority_dict[level]:
                if table not in created and table not in self.with_data \
                        and table not in priority_dict.get(level+1, []):
                    self.create_table(table)
        log.info("Fakeme finished successful \n", datetime.now())

    def resolve_auto_alises(self):
        for table in self.schemas:
            for column in self.schemas[table][0]:
                # column - dict with column description
                alias_column_name = column.alias or column.name
                relative_table = self.table_prefix_in_column_name(alias_column_name)
                if relative_table:
                    matches = self.cfg.matches
                    if self.cfg.tables.get(table):
                        column_settings = self.cfg.tables.get(table).columns.get(column.name, {})
                        if column_settings:
                            matches = column_settings.matches
                        else:
                            matches = self.cfg.tables.get(table).matches
                    if table not in self.rls:
                        self.rls[table] = {}
                    self.rls[table].update({
                        column.name: {
                            'alias': alias_column_name.split('_')[1],
                            'table': relative_table,
                            'matches': matches
                        }
                    })

    def table_prefix_in_column_name(self, column_name: Text) -> Union[Text, None]:
        """ check do we have table name prefix in column name, to get possible auto aliasing """
        table_names = list(self.schemas.keys())

        column_name = column_name.split('_')[0]
        if column_name.endswith('y'):
            table_name = column_name.replace('y', 'ies')
        elif column_name.endswith('s'):
            table_name = column_name + 'es'
        else:
            table_name = column_name + 's'

        if table_name in table_names:
            log.info(f"Found alias with {table_name}")
            return table_name

        return

    def create_table(self, table):
        """ run table creation """
        target_path = os.path.join(self.cfg.path_prefix, "{}.{}".format(table, self.cfg.output.file_format))

        self.schemas[table][1].create_data(
            file_path=target_path,
            with_data=self.with_data,
            chained=self.linked_fields,
            alias_chain=self.rls,
            appends=self.appends,
            prefix=self.cfg.path_prefix,
            cli_path=self.cli_path)

        created.append(table)

    def create_tables_priority_graph(self):
        priority_dict = defaultdict(set)
        # 3
        [priority_dict[3].add(k) for k in self.appends]

        chains_tables = set([k for k in self.rls])

        if self.rls:
            priority_dict[0] = set([table_name for table_name in self.schemas if table_name not in chains_tables])
        else:
            priority_dict[0] = set(self.schemas.keys())

        if self.with_data:
            for data_file in self.with_data:
                priority_dict[0].add(data_file)
        link_dict = {}
        tables_list = [x for x in self.schemas]
        for table in chains_tables:
            self_table = self.rls[table]
            for field in self_table:
                field_dict = self_table[field]
                if field_dict['table'] in tables_list or field_dict['table'] in self.with_data:
                    if not link_dict.get(field_dict['table']):
                        link_dict[field_dict['table']] = set([])
                    link_dict[field_dict['table']].add(table)
        while link_dict:
            for n in range(0, 6):
                table_moved = []
                for table in link_dict:
                    if table in priority_dict[n]:
                        [priority_dict[n+1].add(tb_name) for tb_name in link_dict[table]]
                        table_moved.append(table)
                if table_moved:
                    for table in table_moved:
                        del link_dict[table]
        return priority_dict



import os
import copy

import multiprocessing as mp
from multiprocessing import Queue
from datetime import datetime, timedelta
import pytz
import math
from random import randint
from collections.abc import Iterable
import logging
from typing import Text, List, Dict
from pandas import DataFrame, Series, read_csv, read_json  # noqa F401

from fakeme.fields import FieldRules
from fakeme.values import values_generator, list_generator
from fakeme import config
from fakeme.utils import log

q = Queue()

tmp_prefix = 'tmp_'


class DataGenerator:
    def __init__(self,
                 schema: List,
                 with_data: List = None,
                 chained: Dict = None,
                 table_id: Text = None,
                 alias_chain: Dict = None,
                 appends: Dict = None,
                 cli_path=None,
                 prefix=None
                 ):
        self.schema = schema
        self.chained = chained
        self.table_id = table_id
        self.chains = alias_chain
        self.fr = FieldRules()
        self.chained_df = None
        self.column_df = None
        self.cfg = config.cfg
        self.file_format = self.cfg.output.file_format
        self.appends = appends
        self.with_data = with_data
        self.prefix = prefix if not cli_path else cli_path
        self.table_settings = self.cfg.tables.get(self.table_id, None)
        self.row_numbers = self.table_settings.row_numbers if self.table_settings else self.cfg.row_numbers

    def get_depend_on_file(self):
        """ find depends on other tables (data files)
            TODO: add support for multiple depends on files
        """
        dir_files = []
        if self.chains and self.table_id in self.chains:
            table_chain = self.chains[self.table_id]
            key = list(table_chain.keys())[0]
            file_name = "{}.{}".format(table_chain[key]['table'], self.file_format) if '.' not in table_chain[
                key]['table'] else table_chain[key]['table']
            src_file = os.path.join(self.prefix, file_name)
            if not os.path.isfile(src_file) and self.prefix not in src_file:
                src_file = os.path.join(self.prefix, src_file)
                if not os.path.isfile(src_file):
                    raise ValueError(f'{src_file} is not a file. Current dir {os.getcwd()}')
            dir_files.append(src_file)
        else:
            for item in self.schema:
                if item['name'] in self.chained:
                    chained_tables = [
                        table for table in self.chained[
                            item['name']] if table != self.table_id]
                    [dir_files.append(file_name) if file_name.startswith(table) else None
                     for table in chained_tables for file_name in os.listdir(
                        self.prefix)]
                elif 'all' in self.prefix and item['name'] in self.chains['all']:
                    table_chain = self.chains['all']
                    key = table_chain.keys()[0]
                    if table_chain[key]['table'] != self.table_id:
                        src_file = os.path.join(self.prefix, "{}.{}".format(
                            table_chain[key]['table'], self.file_format))
                        if not os.path.isfile(src_file):
                            raise ValueError
                        dir_files.append(src_file)
        if dir_files:
            log.info("Depend on: {}".format(dir_files))
            return os.path.join(self.prefix, dir_files[0])
        else:
            return []

    def _settings_validation(self, settings):
        raise NotImplementedError

    def resolve_dependencies(self):
        """ find dependencies on different ways and prepare self.chained_df"""
        depend_on_file = None
        if self.appends and self.table_id in self.appends:
            depend_on_file = [os.path.join(self.prefix, "{}.{}".format(
                name, self.file_format)) for name in self.appends[self.table_id]]
        elif self.chained or self.chains:
            depend_on_file = self.get_depend_on_file()
        if depend_on_file:
            if isinstance(depend_on_file, str):
                if self.table_id in os.path.basename(depend_on_file):
                    return tmp_prefix
                self.chained_df = self._read_df_from_file(depend_on_file)
            elif isinstance(depend_on_file, Iterable):
                for file_name in depend_on_file:
                    if self.chained_df is None:
                        self.chained_df = self._read_df_from_file(file_name)
                    else:
                        self.chained_df = self.chained_df.merge(
                            how='outer', right=self._read_df_from_file(file_name))
        return self.chained_df

    def _read_df_from_file(self, depend_on_file):
        """ return correct pandas export method depend on expected file format """
        result = eval(f"read_{self.file_format}(\'{depend_on_file}\')")
        return result

    def _read_values_from_txt(self, path):
        with open(path, 'r') as data_file:
            return [x.replace('\n', '') for x in data_file.readlines()]

    def init_timestamps_values(self):
        # TODO: need to refactor this
        global current_time, last_time

        timezone = pytz.timezone(self.cfg.timezone)
        current_time = datetime.now(timezone).strftime("%Y-%m-%d-%H.%M.%S.%f")
        last_time = (datetime.now(timezone)
                     - timedelta(hours=48)).strftime("%Y-%m-%d-%H.%M.%S.%f")

    def get_data_frame(self):
        """ generate frame with data per table"""
        self.init_timestamps_values()
        logging.info("Start Data Generation")

        data_frame = DataFrame()
        current_obj = copy.deepcopy(self)
        num_cores = mp.cpu_count()
        pool = mp.Pool(num_cores)
        jobs = []
        for num, item in enumerate(self.schema):
            jobs.append(pool.apply_async(column_generation, (current_obj, item)))
            # wait for all jobs to finish
        for job in list(jobs):
            job.get()

        # clean up
        pool.close()

        num = 0
        while not q.empty():
            file_name = q.get()
            with open(file_name, 'r') as column_file:
                column = [line.split('\n')[0] if line != 'None\n' else None for line in column_file.readlines()]
                data_frame.insert(num, column=file_name.split(tmp_prefix)[1],
                                  value=Series(column))
                num += 1
            os.remove(file_name)
        return data_frame

    def get_column_generating_rule(self, field_name):
        """ method to get field rule """
        field_rule = self.fr.rules.loc[self.fr.rules['field']
                                       == field_name].to_dict()
        if not field_rule.get('field'):
            field_rule = {}
        for elem in field_rule:
            field_rule[elem] = field_rule[elem][list(field_rule[elem].keys())[0]]
        return field_rule

    def get_values_from_chained_column(self, _field_name, _matches_k):
        """ get params reversed and matches_k """
        matches = self.chains.get(self.table_id, {}).get(_field_name, {}).get(
            'matches')
        revers = self.chains.get(self.table_id, {}).get('revers', False)
        return matches, revers

    def get_dataframe_column(self, _src_column, _matches_k, _revers):
        """ get column for field from chained dataframe """
        df_column = self.chained_df[[_src_column]]
        splited_index = int(max(df_column.index) * _matches_k)
        if _revers:
            df_column = df_column.reindex(index=df_column.index[::-1])
        df_column = df_column.reindex(index=df_column.index[0:splited_index+1])
        df_column = df_column.to_dict()[_src_column].values()
        return df_column

    def get_column_from_chained(self, field_name, matches_k):
        field_chain = self.chains[self.table_id].get(field_name, {})
        if not field_chain:
            return None
        alias = field_chain.get('alias', None)
        src_column = alias or field_name
        df_column = None
        self.chained_df = self.resolve_dependencies()
        if self.chained_df is not None:
            if isinstance(self.chained_df, str) and self.chained_df == tmp_prefix:
                # mean we depend on column from current ds
                df_column = self._read_values_from_txt(f'{tmp_prefix}{field_name}')
            elif src_column in self.chained_df.columns:
                matches_k, revers = self.get_values_from_chained_column(field_name, matches_k)
                df_column = self.get_dataframe_column(src_column, matches_k, revers)
        return df_column

    def column_generator(self, column_cfg: Dict):
        """ create column with values """
        log.info("Generate column {}".format(column_cfg.name))
        # unique_values - count of unique values in column in this table
        # unique - flag, must be all values unique in this table or not
        column = []
        unique_values = self.row_numbers
        matches_k = self.cfg.matches
        unique = None
        percent_of_nulls = self.cfg.percent_of_nulls

        if self.table_settings:
            column_settings = self.table_settings.columns.get(column_cfg.name, {})
            if column_settings:
                unique_values = column_settings.unique_values or column_settings.row_numbers
                unique = column_settings.unique
                matches_k = column_settings.matches
                percent_of_nulls = column_settings.percent_of_nulls
            else:
                matches_k = self.table_settings.matches
                percent_of_nulls = self.table_settings.percent_of_nulls
        # get field rule
        generating_rule = self.get_column_generating_rule(column_cfg.name)

        # get settings

        if self.table_id in self.chains:
            df_column = self.get_column_from_chained(column_cfg.name, matches_k)
        else:
            df_column = None

        if not unique_values:
            unique_values = self.row_numbers
        if df_column:
            if column_cfg.type == 'LIST':
                # mean we need to create as output lists with values from df_column
                max_number_of_elements = column_cfg.max_number or self.cfg.max_list_values
                min_number_of_elements = column_cfg.min_number or self.cfg.min_list_values
                for _ in range(unique_values):
                    value = list_generator(list(df_column), min_number_of_elements, max_number_of_elements)
                    column.append(value)
            else:
                append_times = int(unique_values / len(df_column)) + 1
                column = list(df_column) * append_times
                column = column[:unique_values]
            if unique:
                column = self.filter_on_unique(column)

        if len(column) < unique_values:
            unique_values = unique_values - len(column)
        else:
            column = column[:unique_values]
            unique_values = 0
        if column_cfg.len and math.isnan(generating_rule['len']) or \
                column_cfg.len and generating_rule['len'] > column_cfg.len:
            generating_rule['len'] = column_cfg.len
        while unique_values:
            value = values_generator(generating_rule, unique)
            column.append(value)
            unique_values -= 1
        total_rows = self.row_numbers - len(column)
        rel_size = total_rows / len(column)
        num_copy = int(rel_size)
        base_column = copy.deepcopy(column)
        for _ in range(num_copy):
            column += base_column
        float_adding = rel_size - num_copy
        column += base_column[:int(len(base_column) * float_adding)]
        if column_cfg.mode:
            nullable = column_cfg.mode == 'nullable'
            if nullable:
                column_len = len(column)
                count_of_nulls = int(column_len * percent_of_nulls)
                for i in range(count_of_nulls):
                    column[randint(0, column_len-1)] = None
        return column

    def filter_on_unique(self, column):
        return list(set(column))


def column_generation(current_obj: DataGenerator,
                      column: dict) -> None:
    file_name = f'{tmp_prefix}{column.name}'
    with open(file_name, 'w+') as column_tmp:
        for line in current_obj.column_generator(column):
            column_tmp.write(str(line))
            column_tmp.write(str('\n'))
    q.put(file_name)



"""  half of imports are used in step of field generation, so this is why imports looks like unused """

import os
import string
import json
import random  # noqa F401
import math
import re  # noqa F401
from uuid import uuid1, uuid4  # noqa F401
from random import randint, random, choice, choices  # noqa F401

from mimesis import Generic, Datetime, Text, Address, Person

from random import getrandbits # noqa F401
# todo: need to add hook with preset data if
#  in rules.json exists generators based on data from 'data' folder
data_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class DataObject:
    pass


languages = ['zh', 'ru', 'en', 'es', 'ar', 'fr']

fake = Generic()
text = Text()
person = Person()
date = Datetime()
address = Address()

current_time = None
last_time = None
countries = DataObject()


def countries_init():
    """
        method to read countries and load data for generator
    :return:
    """
    country_data_file = os.path.join(data_folder, 'countries.json')

    with open(country_data_file, 'r') as countries_file:
        countries_data = json.load(countries_file)
        setattr(countries, 'names', [x['name'] for x in countries_data])
        setattr(countries, 'codes', [x['code'] for x in countries_data])


countries_init()


def random_char(_len):
    return ''.join([choice(string.ascii_letters) for x in range(_len)]).upper()


def values_generator(rule=None, unique=False):
    """ values generator, that eval rules what was defined in 'generator' key  """

    value = None
    if rule and rule['generator']:
        value = eval(rule['generator'])
        if isinstance(rule.get('len', ''), float) and not math.isnan(rule['len']):
            rule['len'] = int(rule['len'])
        if isinstance(rule.get('len', ''), int):
            value = value[0:rule['len']]
        if unique:
            if isinstance(value, str):
                value = value + ' ' + text.word().capitalize()
            elif isinstance(value, int):
                value = value + randint(1000, 10000)
            elif isinstance(value, float):
                value = value + round(random(), 4)
    return value


def list_generator(dataset, min_number_of_elements, max_number_of_elements):
    output = []
    len_dataset = len(dataset)
    count_of_elemnts = randint(min_number_of_elements, max_number_of_elements)
    for i in range(count_of_elemnts):
        output.append(dataset[randint(0, len_dataset-1)])
    return output



import os.path
import json
from copy import copy
from pandas import DataFrame
from fakeme.rules import default_rules
from fakeme.utils import log


class FieldRulesExtractor(object):

    file_name = "rules.json"

    def __init__(self, fields):

        self.fields = self.extract_fields(fields)

    @staticmethod
    def extract_fields(fields):
        _fields = set([])
        fields_with_fixed_rules = [line['field'] for
                                   line in FieldRules.user_rules]
        for table in fields:
            [_fields.add(field) if field not in fields_with_fixed_rules
             else None for field in table[1]]
        return _fields

    def rules_extracts(self):
        field_rules = []
        for field in self.fields:
            for key in default_rules:
                if key in field.lower():
                    field_rule = copy(default_rules[key])
                    break
            else:
                field_rule = copy(default_rules['default'])
            field_rule['field'] = field
            field_rules.append(field_rule)
        [field_rules.append(x) for x in FieldRules.user_rules]
        return field_rules

    def generate_rules(self, remove_existed=True):
        if not remove_existed and os.path.isfile(self.file_name):
            log.info("{} with rules founded in {}".format(self.file_name, os.getcwd()))
        else:
            values_rules_dict = self.rules_extracts()
            with open(self.file_name, 'w+') as outfile:
                json.dump(values_rules_dict, outfile, indent=2)
            log.info("{} with rules for fields was created".format(self.file_name))
        return True

    def get_chains(self, schemas, chains=None):
        """ TODO: add validation for constr "list ON list.grec_cust_id_lsr=hh.grec_cust_id"
            need to add support to chains
        """
        chained = {}
        for table in schemas:
            last_tables = [table_fields for table_fields in schemas
                           if table_fields[0] != schemas[table][0]]
            for second_table in last_tables:
                found_fields = [field['name'] for field in second_table[0] if
                                field in schemas[table][0]]
                if len(found_fields) > 0:
                    for field in found_fields:
                        if chained.get(field):
                            [chained[field].add(
                                name) for name in [table[0], second_table[0]]]
                        else:
                            chained[field] = {table[0], second_table[0]}
        return chained


class FieldRules(object):

    user_rules = []

    def __init__(self):
        try:
            with open(FieldRulesExtractor.file_name, 'r') as json_file:
                list_with_field_rules = json.load(json_file)
        except IOError:
            list_with_field_rules = []

        dict_none_duplicates = {}

        for line in list_with_field_rules:
            dict_none_duplicates[line['field']] = line

        for line in FieldRules.user_rules:
            dict_none_duplicates[line['field']] = line

        final_rules_list = [dict_none_duplicates[line]
                            for line in dict_none_duplicates]
        self.rules = DataFrame.from_records(final_rules_list)



from typing import Text, Union, Any
from logging import getLogger

log = getLogger('_FakeMe_')


def class_to_table_name(cls: Union[Text, Any]):
    """ any mean any class, cls can come as str or as Python class"""
    endings = {'y': lambda x: x[:-1] + "ies", 's': lambda x: x + 'es'}
    if isinstance(cls, str):
        base_name = cls.lower()
    else:
        base_name = cls.__name__.lower()
    table_name = endings.get(base_name[-1], lambda x: x + 's')(base_name)
    return table_name



""" module with command line to run test data generation
    from command line with json config"""
import os.path
import json
import argparse
import inspect
from fakeme.core import Fakeme


def create_cli_parser():
    """ generate command line cli """
    description = "Fakeme - Flexible, customizable data generator for related data"

    parser = argparse.ArgumentParser(description=description,
                                     prog="Fakeme Tables Data Generator")

    parser.add_argument('config',
                        help="path to config file with defined rules "
                             "to data generation, to get example, "
                             "please look at /fakeme/example/example_with_command_line")

    return parser


def run_cli():
    """ main method for fakeme cli """
    parser = create_cli_parser()
    args = parser.parse_args()
    config_path = os.path.expanduser(args.config)
    if not os.getcwd() in config_path:
        config_path = os.path.join(os.getcwd(), config_path)
    if not os.path.isfile(config_path):
        print("Config path {} does not exist. Please, provide "
              "valid path".format(args.config))
        exit(1)
    with open(config_path, 'r') as config:
        conf = json.load(config)
        validate_config(conf)
        Fakeme(cli_path=os.path.dirname(config_path), **conf).run()


def validate_config(conf):
    possible_config_params = inspect.signature(Fakeme.__init__).parameters
    for key, _ in conf.items():
        if key not in possible_config_params:
            print(f"Config contain incorrect values. "
                  f"You can provide only: {possible_config_params} params ")
            exit(1)



""" module that contain config object with default settings """
from typing import Dict, Optional
from pydantic import BaseModel, validator


# current run config instance
cfg = None


class OutputFormat(BaseModel):
    file_format: str = "json"
    config: dict = {}


class ColumnConfig(BaseModel):
    unique_values: bool = False
    row_numbers: Optional[int] = 100
    matches: Optional[int] = 1
    percent_of_nulls: Optional[float] = 0.05


class TableConfig(BaseModel):
    row_numbers: Optional[int] = 100
    columns: Optional[Dict[str, ColumnConfig]] = {}
    percent_of_nulls: Optional[float] = 0.05
    matches: Optional[int] = 1


class Config(BaseModel):
    """ config model """
    row_numbers: int = 100
    matches: float = 1
    timezone: str = "UTC"
    path_prefix: str = "."
    auto_alias: bool = False
    max_list_values: int = 4
    min_list_values: int = 0
    percent_of_nulls: float = 0.05
    output: OutputFormat = OutputFormat(**{'file_format': "json", 'config': {}})
    tables: Optional[Dict[str, TableConfig]] = {}

    @staticmethod
    @validator('row_numbers', 'matches')
    def check_value_above_zero(value):
        if value < 0:
            raise ValueError(f"value must be positive (>0). You set {value}")
        return value



""" classes to work with tables """

import os
from typing import List, Dict, Text
from fakeme.generator import DataGenerator
from fakeme.output import get_writer
from fakeme.schema import SchemaExtractor
from fakeme import config
from fakeme.utils import log


class TableRunner(object):
    """
    `schema_in_obj` arg example:

        [
            {"type": "STRING", "name": "name", "mode": "NULLABLE"},
            {"type": "STRING", "name": "id", "mode": "NULLABLE"}
        ]

    """
    def __init__(self,
                 table_id: Text,
                 schema: List,
                 dataset_id: Text = None
                 ):
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.schema = schema

    def create_data(self,
                    file_path: Text,
                    with_data: List = None,
                    chained: Dict = None,
                    alias_chain: Dict = None,
                    appends: Dict = None,
                    cli_path: Text = None,
                    prefix: Text = None,
                    remove_old: bool = True
                    ):
        """ generate data and save to the file """

        cfg = config.cfg

        table_data = DataGenerator(
            schema=self.schema, with_data=with_data, chained=chained,
            table_id=self.table_id, alias_chain=alias_chain, appends=appends,
            cli_path=cli_path, prefix=prefix).get_data_frame()

        if table_data is None:
            raise Exception("Data Generator failed")
        else:
            target_file_path = self._prepare_path(file_path, remove_old)
            writer = get_writer(cfg.output.file_format)
            writer(table_data, target_file_path, cfg.output.config)

    @staticmethod
    def _prepare_path(file_path, remove_old):
        """ prepare folder and check target file """
        if os.path.isfile(file_path):
            log.info("Founded old file {}".format(file_path))
            if remove_old:
                os.remove(file_path)
                log.info("File {} was removed".format(file_path))
            else:
                raise Exception("Impossible to generate data into file {}. "
                                "File already exist. Please delete file or set "
                                "'remove_old'=True".format(file_path))
        else:
            if not os.path.isdir(os.path.abspath(os.path.dirname(file_path))):
                os.makedirs(os.path.dirname(file_path))
        return file_path


class MultiTableRunner(object):

    def __init__(self, tables: List, rls: Dict):
        """

        :param tables:
        :type tables: list
        :param settings:
        :type settings: dict
        :param prefix:
        :type prefix: str
        """
        self.tables = tables
        self.rls = rls

    def get_fields_and_schemas(self, dump_schema: bool = False):
        """ return list of schemas and tables fields """
        schemas = {}
        fields = []
        for dataset_id, table_id, schema in self.get_values_from_tables_list():

            se = SchemaExtractor(schema=schema,
                                 dataset=dataset_id,
                                 table_id=table_id,
                                 dump_schema=dump_schema)
            schema = se.schema

            if se.rls_founded:
                self.rls.update(se.rls_founded)

            tg = TableRunner(table_id, dataset_id=dataset_id, schema=schema)
            schemas[table_id] = (tg.schema, tg)

        for table_id in schemas:
            fields.append((table_id, [column.name for column in schemas[table_id][0]]))
        return schemas, fields

    def get_values_from_tables_list(self):
        """ this method try to understand that we have in table definition tuple """

        for table_definition in self.tables:

            if isinstance(table_definition, tuple) or isinstance(table_definition, list):
                # we have tuple with table definition
                if len(table_definition) == 3:
                    dataset_id, table_id, schema = table_definition
                elif len(table_definition) == 2:
                    dataset_id = None
                    table_id, schema = table_definition
                elif len(table_definition) == 1:
                    dataset_id, table_id = None, 'undefined'
                    schema = table_definition[0]
                else:
                    raise Exception(f'Invalid table definition in table list, it must contains max 3 values. '
                                    f'You provided: {table_definition}')
            else:
                # we have class definition
                if not getattr(table_definition, '__dict__', None):
                    raise Exception('Table Definition must be the tuple with table properties or class')
                else:
                    dataset_id = table_definition.__module__
                    table_id = table_definition.__name__.lower()
                    schema = table_definition.__annotations__
            yield dataset_id, table_id, schema



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakeme', 'fakeme.cli']

package_data = \
{'': ['*'], 'fakeme': ['data/*']}

install_requires = \
['mimesis>=4.0,<5.0',
 'pandas>=1.0,<2.0',
 'ply>=3.11,<4.0',
 'pydantic>=1.5.1,<2.0.0']

entry_points = \
{'console_scripts': ['fakeme = fakeme.cli:cli']}

setup_kwargs = {
    'name': 'fakeme',
    'version': '0.0.3',
    'description': 'Relative Data Generator: generate tables with data, that depend on each other',
    'long_description': 'Fakeme \n=======\n\nData Generator for Chained and Relative Data\n\n|badge1| |badge2| |badge3|\n\n.. |badge1| image:: https://img.shields.io/pypi/pyversions/fakeme \n.. |badge2| image:: https://img.shields.io/pypi/v/fakeme\n.. |badge3| image:: https://travis-ci.com/xnuinside/fakeme.svg?branch=master\n\nDocumentation in process: https://fakeme.readthedocs.io/en/latest/ \n\nUse under the hood at current time:\n\n     - Mimesis (like one of the generators for fields values)\n     - Pandas (like main instrument to combain data in tables (frames) \n     - Standart Python Library\n\n\nSupport Python 3.7\n\nWhat it does not do?\n=========================\n\nThis is not random value generated library - for this exist pretty cool lib Mimesis and another. \n\n\nFor what it\n=========================\n\n**Fakeme** oriented on generation data that depend on values in another tables/datasets.\nData, that knitted together as real. \n\n**Fakeme** can help you if you want to generate several tables, that must contains in columns values, \nthat you will use like key for join.\n\nFor example, *user_data* table has field *user_id* and *users* table contains list of users with column id. \nYou want join it on user_id = id.\n\n**Fakeme** will generate for you 2 tables with same values in those 2 columns. \n\nIt does not matter to have columns with same name you can define dependencies between tables with alias names. \n \n \nWhat you can to do\n=========================\n\n1. Define that fields in your datasets must contain similar values\n\n2. You can set up how much values must intersect, for example, you want to emulate data for email validation pipeline - \nyou have one dataset with *incoming* messages  and you need to find all emails that was not added previously in your *contacts* table.\n\nSo you will have incoming messages table, that contains, for example only 70% of emails that exist in contacts table. \n\n3. You can use multiply columns as a key (dependency) in another column, for example, \n*player_final_report* must contains for each player same values as in other tables, for example, you have *player* table\nwith players details and *in_game_player_activity* with all player activities for some test reasons it\'s critical\nto you generate *player_final_report* with 1-to-1 data from other 2 tables.\n \n4. Union tables. You can generate tables that contains all rows from another tables. \n\n5. You can define your own generator for fields on Python.\n\n6. You can define your own output format\n\n\nExamples\n=========================\n\n   You can find usage examples in \'fakeme/examples/\' folder.\n\n    Example from fakeme/examples/generate_data_related_to_existed_files:\n\n.. code-block:: python\n\n    from fakeme import Fakeme\n\n    # to use some existed data you should provide with_data argument -\n    # and put inside list with the paths to the file with data\n\n    # data file must be in same format as .json or csv output of fakeme.\n    # so it must be [{"column_name": value, "column_name2": value2 ..},\n    #   {"column_name" : value, "column_name2": value2 ..}..]\n    # Please check example in countries.json\n\n    cities_schema = [{"name": "name"},\n                     {"name": "country_id"},\n                     {"name": "id"}]\n\n    # all fields are strings - so I don\'t define type, because String will be used as default type for the column\n\n    tables_list = [(\'cities\', cities_schema)]\n\n    Fakeme(\n        tables=tables_list,\n        with_data=[\'countries.json\'],\n        rls={\'cities\': {  # this mean: for table \'cities\'\n            \'country_id\': {  # and column \'country_id\' in table \'cities\'\n                \'table\': \'countries.json\',   # please take data from data  in countries.json\n                \'alias\': \'id\',  # with alias name \'id\'\n                \'matches\': 1  # and I want all values in country_id must be from countries.id column, all.\n            }\n        }},\n        settings={\'row_numbers\': 1300}  # we want to have several cities for each country,\n                                        # so we need to have more rows,\n    ).run()\n\n    # run and you will see a list of cities, that generated with same ids as in countries.json\n\n\nDocs: https://fakeme.readthedocs.io/en/latest/',
    'author': 'xnuinside',
    'author_email': 'xnuinside@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/xnuinside/fakeme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
    class to write DataFrame in output format,
    that you need, can be customised by using

    import fakeme.output
    fakeme.output.user_defined_writers['your_format_name'] = python_callable

"""
import json
from copy import deepcopy
from typing import Text, Dict, Callable
from pandas import ExcelWriter, DataFrame


default_format_setting = {
    'to_json': {
        'orient': 'records',
        'lines': False
    },
    'to_csv': {
        'header': True,
        'index': False,
        'line_terminator': '\n'
    }
}


def to_excel(df: DataFrame) -> Callable:
    to_method = getattr(df, 'to_excel')

    def excel_method(*args, **kwargs):
        path = kwargs['path']
        del kwargs['path']
        return to_method(ExcelWriter(path), *args, **kwargs)
    return excel_method


def output_writer(method_name: Text) -> Callable:
    def format_writer_with_defaults(df: DataFrame, path: Text, params: Dict = None):
        if not params:
            params = {}
        combined_params = deepcopy(default_format_setting[method_name])
        combined_params.update(params)
        return getattr(df, method_name)(path, **combined_params)
    return format_writer_with_defaults


def json_with_indent() -> Callable:
    def json_with_indent_inner(df: DataFrame, path: Text, params: Dict = None) -> None:
        if not params:
            params = {}
        combined_params = deepcopy(default_format_setting['to_json'])
        combined_params.update(params)
        with open(path, 'w+') as file_to_write:
            json.dump(json.loads(df.to_json(**combined_params)), file_to_write, indent=1)
    return json_with_indent_inner


supported_formats = {'json': json_with_indent(),
                     'csv': output_writer('to_csv'),
                     'feather': lambda df: getattr(df, 'to_feather'),
                     'md': lambda df: getattr(df, 'to_markdown'),
                     'parquet': lambda df: getattr(df, 'to_parquet'),
                     'html': lambda df: getattr(df, 'to_html'),
                     'excel': to_excel,
                     'hdf': lambda df: getattr(df, 'to_hdf'),
                     'pickle': lambda df: getattr(df, 'to_pickle')}

user_defined_writers = {}


def get_writer(format_key: Text) -> Callable:
    supported_formats.update(user_defined_writers)
    writer = supported_formats.get(format_key)
    if not writer:
        raise Exception(f'Output format {format_key} not exist. '
                        f'Please, provide correct name or add your custom output writer'
                        f'with OutputWriter.user_defined_writers[\'your_format_name\'] = python_callable')
    return writer



import json
import os
from typing import Text, Optional, Dict, List, Union
from fakeme.utils import class_to_table_name
from fakeme import config
from pydantic import BaseModel, validator


supported_types = ['STRING', 'INTEGER', 'FLOAT', 'LIST']


types_map = {

    'str': 'string',
    'int': 'integer',
    'float': 'float',
    'list': 'list',
    'dict': 'dict',
    'datetime': 'string'
}


class Column(BaseModel):
    name: str
    type: str = str
    len: Optional[int] = None
    mode: Optional[str] = None
    max_number: Optional[int] = None
    min_number: Optional[int] = None
    alias: Optional[str] = None

    @staticmethod
    @validator('mode')
    def set_mode_lower(value):
        return value.lower()

    @staticmethod
    @validator('type')
    def set_type_upper(value):
        return value.upper()

    # todo: add type validation
    """
    def schema_validation(self, schema):
        for item in schema:
            _type = item.type
            self.type_validation(_type)
        return schema

    @staticmethod
    def type_validation(type_name):
        if type_name.upper() not in supported_types:
            raise ValueError

    """


class SchemaExtractor(object):

    def __init__(self,
                 schema: Text = None,
                 dataset: Text = None,
                 table_id: Text = "",
                 dump_schema: bool = True):
        self.schema = self.get_schema(schema)
        self.dataset = dataset or self.extract_folder()
        self.table_id = table_id
        self.dump_schema = dump_schema
        self.rls_founded = {}

    @staticmethod
    def _validate_path(path: Text):
        """ check schema path exists, try to get correct path """
        path = os.path.expanduser(path)
        if not os.path.isfile(path):
            os.chdir(config.cfg.path_prefix)
            target_path = os.path.abspath(path)
        else:
            target_path = path
        if not os.path.isfile(target_path):
            raise ValueError("File with schema does not exist {}".format(path))
        return target_path

    def get_schema_from_file(self, schema_path: Text):
        if isinstance(schema_path, str):
            schema_path = self._validate_path(schema_path)
            if schema_path.endswith('.json'):
                with open(schema_path, 'r') as schema_file:
                    schema = [Column(**column) for column in json.load(schema_file)]
            else:
                raise NotImplementedError('Supports only `.json` format')
        return schema

    def get_schema(self, schema: Union[Dict, List, Text]):
        if isinstance(schema, str):
            schema = self.get_schema_from_file(schema)
        else:
            schema = self.get_schema_from_python_obj(schema)
        return schema

    def extract_folder(self):
        if '/' in self.schema:
            splited = self.schema.split('/')
            folder, _ = splited[(len(splited) - 2):]
            if folder == 'schemas':
                folder = ''
        else:
            folder = '.'
        return folder

    def get_schema_from_python_obj(self, schema: Union[Dict, List]):
        """ get schema in BQ format """
        # we have a dict from class annotations
        if isinstance(schema, dict):
            schema_fields = []
            for key in schema:
                if isinstance(schema[key], str):
                    # mean we have type definition
                    _type = schema[key]
                    if _type and _type.__module__ == self.dataset:
                        # mean class defined in same module
                        # need to set dependency
                        if not self.rls_founded.get(self.table_id):
                            self.rls_founded[self.table_id] = {}
                        self.rls_founded[self.table_id][class_to_table_name(key)] = {
                                'alias': 'id',
                                'table':  class_to_table_name(_type)
                            }
                        _type = None
                    else:
                        _type = self.map_type(_type)
                    field_details = {'name': key, 'type': _type}
                else:
                    if not schema[key].get('name'):
                        schema[key]['name'] = key
                    field_details = schema[key]
                field = Column(**field_details)
                schema_fields.append(field)
            schema = schema_fields
        elif isinstance(schema, list):
            schema = [Column(**field) for field in schema]
            return schema
        return schema

    @staticmethod
    def map_type(_type):
        """ method to map input types to the types that understandable by generator"""
        # in any unknown situation - use string :D
        mapped_type = 'string'
        if isinstance(_type, type):
            name = _type.__name__
            mapped_type = types_map[name]
        return mapped_type
